@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 1.2
+;; Version: 1.3
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.4.4"))
 ;; Homepage: https://github.com/minad/org-modern
 ;; Keywords: outlines, hypermedia, text
@@ -53,11 +53,26 @@ If set to `auto' the border width is computed based on the `line-spacing'.
 A value between 0.1 and 0.4 of `line-spacing' is recommended."
   :type '(choice (const nil) (const auto) integer))
 
-(defcustom org-modern-star "◉○◈◇✳"
-  "Replacement strings for headline stars for each level.
-Set to nil to disable styling the headlines."
+(defcustom org-modern-star 'fold
+  "Style heading stars.
+Can be nil, fold or replace.  See `org-modern-fold-stars' and
+`org-modern-replace-stars' for the respective configurations."
+  :type '(choice (const :tag "No styling" nil)
+                 (const :tag "Folding indicators" fold)
+                 (const :tag "Replace" replace)))
+
+(defcustom org-modern-replace-stars "◉○◈◇✳"
+  "Replacement strings for headline stars for each level."
   :type '(choice string (repeat string)))
 
+(defcustom org-modern-fold-stars
+  '(("▶" . "▼") ("▷" . "▽") ("⯈" . "⯆") ("▹" . "▿") ("▸" . "▾"))
+  "Folding indicators for headings.
+Replace headings' stars with an indicator showing whether its
+tree is folded or expanded."
+  :type '(repeat (cons (string :tag "Folded")
+                       (string :tag "Expanded"))))
+
 (defcustom org-modern-hide-stars 'leading
   "Changes the displays of the stars.
 Can be leading, t, or a string/character replacement for each
@@ -138,12 +153,24 @@ replacement expression, e.g., a string."
 
 (defcustom org-modern-todo-faces nil
   "Faces for todo keywords.
-This is an alist, with todo keywords in the car
-and faces in the cdr.  Example:
+This is an alist, with todo keywords in the car and faces in the
+cdr.  Example:
 
   (setq org-modern-todo-faces
-    (quote ((\"TODO\" :background \"red\"
-                    :foreground \"yellow\"))))"
+    (quote ((\"TODO\" :background \"red\" :foreground \"yellow\"))))"
+  :type '(repeat
+          (cons (choice
+                  (string :tag "Keyword")
+                  (const :tag "Default" t))
+                (sexp :tag "Face   "))))
+
+(defcustom org-modern-tag-faces nil
+  "Faces for tags keywords.
+This is an alist, with tag the car and faces in the cdr.
+Example:
+
+  (setq org-modern-tag-faces
+    (quote ((\"emacs\" :background \"red\" :foreground \"yellow\"))))"
   :type '(repeat
           (cons (choice
                   (string :tag "Keyword")
@@ -231,14 +258,11 @@ references."
   "Prettify radio link targets, e.g., <<<radio>>>."
   :type '(choice (const nil) (list string boolean string)))
 
-(defcustom org-modern-statistics t
-  "Prettify todo statistics."
-  :type 'boolean)
-
-(defcustom org-modern-progress "○◔◑◕●"
-  "Add a progress indicator to the todo statistics.
-Set to nil to disable the indicator."
-  :type '(choice string (repeat string)))
+(defcustom org-modern-progress 12
+  "Width in characters to draw progress bars.
+Set to nil to disable bars."
+  :type '(choice (const :tag "Disable progress bar" nil)
+                 (natnum :tag "Bar width")))
 
 (defgroup org-modern-faces nil
   "Faces used by `org-modern'."
@@ -264,6 +288,17 @@ the font.")
   '((t :height 0.8 :weight light))
   "Face used for block keywords.")
 
+(defface org-modern-progress-complete
+  '((((background light))
+     :background "gray35" :foreground "white")
+    (t :background "gray75" :foreground "black"))
+  "Face used for completed section of progress bars (colors only).")
+
+(defface org-modern-progress-incomplete
+  '((((background light)) :background "gray90" :foreground "black")
+    (t :background "gray20" :foreground "white"))
+  "Face used for incomplete section of progress bars (colors only).")
+
 (defface org-modern-tag
   '((default :inherit (secondary-selection org-modern-label))
     (((background light)) :foreground "black")
@@ -282,7 +317,7 @@ the font.")
   '((default :inherit org-modern-label)
     (((background light)) :background "gray90" :foreground "black")
     (t :background "gray20" :foreground "white"))
-  "Face used for done labels.")
+  "Default face used for done labels.")
 
 (defface org-modern-todo
   ;; `:inverse-video' to use todo foreground as label background
@@ -296,12 +331,10 @@ the font.")
        :weight semibold :inverse-video t))
   "Default face used for priority labels.")
 
-(defface org-modern-statistics
-  '((t :inherit org-modern-done))
-  "Face used for todo statistics labels.")
-
 (defface org-modern-date-active
-  '((t :inherit org-modern-done))
+  '((default :inherit org-modern-label)
+    (((background light)) :background "gray90" :foreground "black")
+    (t :background "gray20" :foreground "white"))
   "Face used for active date labels.")
 
 (defface org-modern-time-active
@@ -333,10 +366,10 @@ the font.")
   "Face used for horizontal ruler.")
 
 (defvar-local org-modern--font-lock-keywords nil)
-(defvar-local org-modern--star-cache nil)
+(defvar-local org-modern--folded-star-cache nil)
+(defvar-local org-modern--expanded-star-cache nil)
 (defvar-local org-modern--hide-stars-cache nil)
 (defvar-local org-modern--checkbox-cache nil)
-(defvar-local org-modern--progress-cache nil)
 (defvar-local org-modern--table-sp-width 0)
 (defconst org-modern--table-overline '(:overline t))
 (defconst org-modern--table-sp '((space :width (org-modern--table-sp-width))
@@ -355,7 +388,8 @@ the font.")
   (let ((beg (match-beginning 0))
         (end (match-end 0))
         (rep (and (listp org-modern-keyword)
-                  (cdr (assoc (downcase (match-string 2)) org-modern-keyword)))))
+                  (cdr (assoc (downcase (match-string-no-properties 2))
+                              org-modern-keyword)))))
     (unless rep
       (setq rep (cdr (assq t org-modern-keyword)) end (match-end 1)))
     (pcase rep
@@ -381,18 +415,31 @@ the font.")
          'org-modern-priority)))))
 
 (defun org-modern--progress ()
-  "Prettify headline todo progress."
-  (put-text-property
-   (match-beginning 2) (match-end 2) 'display
-   (aref org-modern--progress-cache
-         (floor
-          (* (1- (length org-modern--progress-cache))
-             (if (match-beginning 3)
-                 (* 0.01 (string-to-number (match-string 3)))
-               (let ((q (string-to-number (match-string 5))))
-                 (if (= q 0)
-                     1.0
-                   (/ (* 1.0 (string-to-number (match-string 4))) q)))))))))
+  "Prettify progress as color-coded bar."
+  (let* ((beg (match-beginning 1))
+         (end (match-end 1))
+         (val (min 1.0
+                   (if (match-beginning 2)
+                       (* 0.01 (string-to-number (match-string-no-properties 2)))
+                     (let ((q (string-to-number (match-string-no-properties 4))))
+                       (if (= q 0)
+                           1.0
+                         (/ (* 1.0 (string-to-number (match-string-no-properties 3))) q))))))
+         (w org-modern-progress)
+         (complete (floor (* w val)))
+         (w0 (- end beg 2))
+         (w1 (/ (- w w0) 2))
+         (bar (concat (make-string w1 ?\s)
+                      (buffer-substring-no-properties (1+ beg) (1- end))
+                      (make-string (- w w1 w0) ?\s))))
+    (put-text-property 0 complete 'face 'org-modern-progress-complete bar)
+    (put-text-property complete w 'face 'org-modern-progress-incomplete bar)
+    (put-text-property beg end 'face 'org-modern-label)
+    (put-text-property beg (1+ beg) 'display (substring bar 0 w1))
+    (put-text-property (1- end) end 'display (substring bar (+ w1 w0) w))
+    (dotimes (i w0)
+      (put-text-property (+ 1 beg i) (+ 2 beg i)
+                         'display (substring bar (+ w1 i) (+ w1 i 1))))))
 
 (defun org-modern--tag ()
   "Prettify headline tags."
@@ -411,13 +458,19 @@ the font.")
                                (format #(" %c" 1 3 (cursor t)) (char-after colon-end)))
             (put-text-property (1- cbeg) cbeg 'display
                                (string (char-before cbeg) ?\s))
-            (put-text-property colon-end cbeg 'face 'org-modern-tag))
+            (put-text-property
+             colon-end cbeg 'face
+             (if-let ((faces org-modern-tag-faces)
+                      (face (or (cdr (assoc (buffer-substring-no-properties colon-end cbeg) faces))
+                                (cdr (assq t faces)))))
+                 `(:inherit (,face org-modern-tag))
+               'org-modern-tag)))
           (add-text-properties cbeg cend colon-props)
           (setq colon-beg cbeg colon-end cend))))))
 
 (defun org-modern--todo ()
   "Prettify headline todo keywords."
-  (let ((todo (match-string 1))
+  (let ((todo (match-string-no-properties 1))
         (beg (match-beginning 1))
         (end (match-end 1)))
     (put-text-property beg (1+ beg) 'display
@@ -488,8 +541,19 @@ the font.")
       (put-text-property
        (if (eq org-modern-hide-stars 'leading) beg end)
        (1+ end) 'display
-       (aref org-modern--star-cache
-             (min (1- (length org-modern--star-cache)) level))))))
+       (let ((cache (if (and org-modern--folded-star-cache
+                             (org-invisible-p (pos-eol)))
+                        org-modern--folded-star-cache
+                      org-modern--expanded-star-cache)))
+         (aref cache (min (1- (length cache)) level)))))))
+
+(defun org-modern--cycle (state)
+  "Flush font-lock for buffer or line at point for `org-cycle-hook'.
+When STATE is `overview', `contents', or `all', flush for the
+whole buffer; otherwise, for the line at point."
+  (pcase state
+    ((or 'overview 'contents 'all) (font-lock-flush))
+    (_ (font-lock-flush (pos-bol) (pos-eol)))))
 
 (defun org-modern--table ()
   "Prettify vertical table lines."
@@ -539,7 +603,7 @@ the font.")
          (beg-name (match-beginning 3))
          (end-name (match-end 3))
          (names (and (listp org-modern-block-name) org-modern-block-name))
-         (rep (cdr (assoc (downcase (match-string 3)) names)))
+         (rep (cdr (assoc (downcase (match-string-no-properties 3)) names)))
          (fringe (and org-modern-block-fringe (not (bound-and-true-p org-indent-mode)))))
     (unless rep
       (setq rep (cdr (assq t names)) end-rep beg-name))
@@ -607,7 +671,7 @@ the font.")
    :box
    (when org-modern-label-border
      (let ((border (if (eq org-modern-label-border 'auto)
-                       (max 3 (cond
+                       (max 2 (cond
                                ((integerp line-spacing)
                                 line-spacing)
                                ((floatp line-spacing)
@@ -657,7 +721,7 @@ the font.")
      `(("^\\*+.*? \\(\\(\\[\\)#.\\(\\]\\)\\) "
         (1 (org-modern--priority)))))
    (when org-modern-todo
-     `((,(format "^\\*+ +%s " (regexp-opt org-todo-keywords-1 t))
+     `((,(format "^\\*+ +%s\\(?: \\|$\\)" (regexp-opt org-todo-keywords-1 t))
         (0 (org-modern--todo)))))
    (when org-modern-checkbox
      `((,org-list-full-item-re
@@ -708,11 +772,9 @@ the font.")
        ("<[^>]+>\\(-\\)\\(-\\)<[^>]+>\\|\\[[^]]+\\]\\(?1:-\\)\\(?2:-\\)\\[[^]]+\\]"
         (1 '(face org-modern-label display #("  " 1 2 (face (:strike-through t) cursor t))) t)
         (2 '(face org-modern-label display #("  " 0 1 (face (:strike-through t)))) t))))
-   (when org-modern-statistics
-     `((" \\(\\(\\[\\)\\(?:\\([0-9]+\\)%\\|\\([0-9]+\\)/\\([0-9]+\\)\\)\\(\\]\\)\\)"
-        (1 '(face org-modern-statistics) t)
-        (2 ,(if org-modern-progress '(org-modern--progress) ''(face nil display " ")))
-        (6 '(face nil display " ")))))
+   (when (integerp org-modern-progress)
+     `((" \\(\\[\\(?:\\([0-9]+\\)%\\|\\([0-9]+\\)/\\([0-9]+\\)\\)]\\)"
+        (0 (org-modern--progress)))))
    (when org-modern-tag
      `((,(concat "^\\*+.*?\\( \\)\\(:\\(?:" org-tag-re ":\\)+\\)[ \t]*$")
         (0 (org-modern--tag)))))
@@ -767,16 +829,18 @@ the font.")
    (org-modern-mode
     (add-to-invisibility-spec 'org-modern)
     (setq
-     org-modern--star-cache
-     (vconcat (mapcar #'org-modern--symbol org-modern-star))
+     org-modern--folded-star-cache
+     (and (eq org-modern-star 'fold)
+          (vconcat (mapcar #'org-modern--symbol (mapcar #'car org-modern-fold-stars))))
+     org-modern--expanded-star-cache
+     (and org-modern-star
+          (vconcat (mapcar #'org-modern--symbol (if (eq org-modern-star 'fold)
+                                                    (mapcar #'cdr org-modern-fold-stars)
+                                                  org-modern-replace-stars))))
      org-modern--hide-stars-cache
      (and (char-or-string-p org-modern-hide-stars)
           (list (org-modern--symbol org-modern-hide-stars)
                 (org-modern--symbol org-modern-hide-stars)))
-     org-modern--progress-cache
-     (vconcat (mapcar
-               (lambda (x) (concat " " (org-modern--symbol x) " "))
-               org-modern-progress))
      org-modern--checkbox-cache
      (mapcar (pcase-lambda (`(,k . ,v)) (cons k (org-modern--symbol v)))
              org-modern-checkbox)
@@ -789,6 +853,8 @@ the font.")
     (add-hook 'pre-redisplay-functions #'org-modern--pre-redisplay nil 'local)
     (add-hook 'org-after-promote-entry-hook #'org-modern--unfontify-line nil 'local)
     (add-hook 'org-after-demote-entry-hook #'org-modern--unfontify-line nil 'local)
+    (when (eq org-modern-star 'fold)
+      (add-hook 'org-cycle-hook #'org-modern--cycle nil 'local))
     (org-modern--update-label-face)
     (org-modern--update-fringe-bitmaps))
    (t
@@ -798,7 +864,9 @@ the font.")
     (setq-local font-lock-unfontify-region-function #'org-unfontify-region)
     (remove-hook 'pre-redisplay-functions #'org-modern--pre-redisplay 'local)
     (remove-hook 'org-after-promote-entry-hook #'org-modern--unfontify-line 'local)
-    (remove-hook 'org-after-demote-entry-hook #'org-modern--unfontify-line 'local)))
+    (remove-hook 'org-after-demote-entry-hook #'org-modern--unfontify-line 'local)
+    (when (eq org-modern-star 'fold)
+      (remove-hook 'org-cycle-hook #'org-modern--cycle 'local))))
   (without-restriction
     (with-silent-modifications
       (org-modern--unfontify (point-min) (point-max)))
